/*
    <one line to give the library's name and an idea of what it does.>
    Copyright (C) 2012  Christian Mollekopf <chrigi_1@fastmail.fm>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Affero General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include "fbgeneratorjob.h"
#include "authenticationjob.h"
#include "generatefbjob.h"
#include "settings.h"
#include "jobs/probekolabserverjob.h"
#include <jobs/messagemodifyjob.h>
#include <jobs/fetchmessagesjob.h>
#include <jobs/sequentialcompositejob.h>
#include <commonconversion.h>

#include <kolabdefinitions.h>
#include <kolabobject.h>
#include <errorhandler.h>
#include <kdebug.h>
#include <kimap/logoutjob.h>
#include <kimap/expungejob.h>


FBGeneratorJob::FBGeneratorJob(const QString& hostName, qint16 port, const QString& username, QObject* parent)
:   KolabJob(hostName, port, username, parent),
    mOldImapUid(-1)
{

}

QStringList FBGeneratorJob::requiredFolders()
{
    return QStringList() << KOLAB_FOLDER_TYPE_FREEBUSY;
}


void FBGeneratorJob::startWork(ProbeKolabServerJob* capabilitiesJob)
{
    if (capabilitiesJob->kolabFolders().values(KOLAB_FOLDER_TYPE_FREEBUSY).isEmpty()) {
        kWarning() << "no freebusy folder found";
        setError(KJob::UserDefinedError);
        logout();
        return;
    }
    mFreebusyFolder = capabilitiesJob->kolabFolders().values(KOLAB_FOLDER_TYPE_FREEBUSY).first();
    mEventFolders = capabilitiesJob->kolabFolders().values(KOLAB_FOLDER_TYPE_EVENT);
    if (mEventFolders.isEmpty()) {
        kWarning() << "no event folders available";
        logout();
        return;
    }

    FetchMessagesJob *fetchJob = new FetchMessagesJob(mFreebusyFolder, mSession, this);
    connect( fetchJob, SIGNAL(result(KJob*)), this, SLOT(onFetchFBDone(KJob*)) );
    fetchJob->start();
}


void FBGeneratorJob::onFetchFBDone(KJob *job)
{
    if ( job->error() ) {
        kWarning() << job->errorString();
        setError(KJob::UserDefinedError);
        logout();
        return;
    }
    FetchMessagesJob *fetchJob = qobject_cast<FetchMessagesJob*>( job );
    Q_ASSERT(fetchJob);

    kDebug() << fetchJob->getMessages().size();
    foreach (const KMime::Message::Ptr &msg, fetchJob->getMessages()) {
        kDebug() << "existing fb object " << msg->subject()->asUnicodeString();
    }

    KDateTime startOfTimeFrame = KDateTime::currentUtcDateTime();
    const KDateTime endOfTimeFrame = startOfTimeFrame.addDays(Settings::instance().getTimeframe());
    int threshold = Settings::instance().getThreshold();

    /*
     * The f/b object to update is identified based on a Mime-Header
     *
     * For the local server we're going to use the Settings::getServerUri()
     */
    qint16 port;
    const QString localUri = Settings::instance().getServerUri(port).toUtf8();
    int count = 0;
    bool upToDate = false;
    foreach (const KMime::Message::Ptr &msg, fetchJob->getMessages()) {
        KMime::Headers::Base *xOriginHeader = msg->getHeaderByType(X_ORIGIN_HEADER);
        if (xOriginHeader) qDebug() << "existing fb object " << xOriginHeader->asUnicodeString() << localUri;
        if (xOriginHeader && (xOriginHeader->asUnicodeString() == localUri)) {

            const Kolab::Freebusy &oldFB = Kolab::KolabObjectReader(msg).getFreebusy();
            const KDateTime dtstamp = Kolab::Conversion::toDate(oldFB.timestamp());
            const KDateTime start = Kolab::Conversion::toDate(oldFB.start());
            const KDateTime end = Kolab::Conversion::toDate(oldFB.end());

            qDebug() << startOfTimeFrame << start << endOfTimeFrame << end << threshold;
            if (!(startOfTimeFrame < start || endOfTimeFrame > end.addDays(threshold))) {
                qDebug() << "within threshold, skipping";
                upToDate = true;
                continue;
            }
            
            if (mOldImapUid < 0) {
                mOldImapUid = fetchJob->getImapUid(msg);
                mOldFlags = fetchJob->getFlags(msg);
            }
            count++;
        }
    }
    if (count > 1) {
        Warning() << "multiple old fb objects detected this shouldn't ever happen";
        //TODO cleanup
    }
    //FIXME we also need to check that no new events are existing, otherwise this check prevents us from generating up-to date information
/*    if (upToDate) {
        Debug() << "fb objects are up-to-date, nothing to do";
        logout();
        return;
    }
*/

//     kDebug() << mEventFolder;
    GenerateFBJob *fbJob = new GenerateFBJob(mEventFolders, startOfTimeFrame, endOfTimeFrame, mSession, this);
    QObject::connect(fbJob, SIGNAL(result(KJob*)), this, SLOT(onGenerateFBDone(KJob*)));
    fbJob->start();
}

void FBGeneratorJob::onGenerateFBDone(KJob *job)
{
    if ( job->error() ) {
        Warning() << "Error " << job->errorString();
        logout();
        return;
    }
    GenerateFBJob *fbJob = qobject_cast<GenerateFBJob*>( job );
    Q_ASSERT(fbJob);

    Kolab::Freebusy fb = fbJob->getFreebusy();
    if (!fb.isValid()) {
        Debug() << "did not generate a fb object";
        //TODO generate empty fb object to mark time as free
        logout();
        return;
    }
    if (!fb.organizer().isValid()) {
//         fb.setOrganizer(/*mUserName, mUserName+QLatin1Char('@')+mHostName.remove("imap.")*/);
        Warning() << "no valid organizer";
        setError(KJob::UserDefinedError);
        logout();
        return;
    }

    KMime::Message::Ptr message = Kolab::KolabObjectWriter::writeFreebusy(fb);
    if (Kolab::ErrorHandler::instance().error() > Kolab::ErrorHandler::Debug) {
        Warning() << "Error: " << Kolab::ErrorHandler::instance().errorMessage();
        setError(KJob::UserDefinedError);
        logout();
        return;
    }

    qint16 port;
    message->appendHeader( new KMime::Headers::Generic( X_ORIGIN_HEADER, message.get(), Settings::instance().getServerUri(port).toUtf8(), "utf-8" ) );
    message->assemble();
    
//     kDebug() << message->encodedContent();
    SequentialCompositeJob *seqJob = new SequentialCompositeJob(this);
    seqJob->addSubjob(new MessageModifyJob(message, mFreebusyFolder, mOldFlags, mOldImapUid, mSession, this));
    seqJob->addSubjob(new KIMAP::ExpungeJob(mSession));
    seqJob->addSubjob(new KIMAP::LogoutJob(mSession));
    QObject::connect(seqJob, SIGNAL(result(KJob*)), this, SLOT(onModDone(KJob*)));
    seqJob->start();
}

void FBGeneratorJob::onModDone(KJob *job)
{
    if ( job->error() ) {
        kWarning() << job->errorString();
    }
    logout();
}

