/*
 * Copyright (C) 2012  Christian Mollekopf <mollekopf@kolabsys.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "fetchmessagesjob.h"
#include <kimap/fetchjob.h>
#include <kimap/selectjob.h>

#include <freebusy.h>
#include <kolabobject.h>
#include <errorhandler.h>
#include <kcalconversion.h>

FetchMessagesJob::FetchMessagesJob(const QString &folder, KIMAP::Session *session, QObject* parent)
:   KJob(parent),
    mSession(session),
    mFolder(folder)
{

}

void FetchMessagesJob::start()
{
    Debug() << "Fetching messages from Mailbox....... " << mFolder;
    KIMAP::SelectJob *select = new KIMAP::SelectJob( mSession );
    select->setMailBox( mFolder );
    connect( select, SIGNAL(result(KJob*)), this, SLOT(onSelectDone(KJob*)) );
    select->start();
}

void FetchMessagesJob::onSelectDone(KJob *job)
{
    if ( job->error() ) {
        Warning() << job->errorString();
        setError(KJob::UserDefinedError);
        emitResult();
        return;
    }

    KIMAP::SelectJob *select = qobject_cast<KIMAP::SelectJob*>( job );
    Q_ASSERT(select);

    const int messageCount = select->messageCount();

    if (messageCount <= 0) {
        //empty, nothing to do
        Debug() << "no messages available, nothing to do";
        emitResult();
        return;
    }

    KIMAP::FetchJob::FetchScope scope;
    scope.mode = KIMAP::FetchJob::FetchScope::Full;

    KIMAP::FetchJob *fetch = new KIMAP::FetchJob( mSession );
    fetch->setSequenceSet( KIMAP::ImapSet( 1, messageCount ) );
    fetch->setScope( scope );
    connect( fetch, SIGNAL( headersReceived( QString, QMap<qint64, qint64>, QMap<qint64, qint64>,
                                                QMap<qint64, KIMAP::MessageFlags>, QMap<qint64, KIMAP::MessagePtr> ) ),
                this, SLOT( onHeadersReceived( QString, QMap<qint64, qint64>, QMap<qint64, qint64>,
                                            QMap<qint64, KIMAP::MessageFlags>, QMap<qint64, KIMAP::MessagePtr> ) ) );
    connect( fetch, SIGNAL(result(KJob*)),
                this, SLOT(onHeadersFetchDone(KJob*)) );
    fetch->start();
}

void FetchMessagesJob::onHeadersReceived( const QString &mailBox,
                                           const QMap<qint64, qint64> &uids,
                                           const QMap<qint64, qint64> &sizes,
                                           const QMap<qint64, KIMAP::MessageFlags> &flags,
                                           const QMap<qint64, KIMAP::MessagePtr> &messages )
{
//     Debug() << mailBox;
    foreach (qint64 key, uids.keys()) {
        Q_ASSERT(messages.contains(key));
        mMessages.insert(key, messages.value(key));
        Q_ASSERT(flags.contains(key));
        mFlags.insert(key, flags.value(key));
        mUids.insert(key, uids.value(key));
    }
}

void FetchMessagesJob::onHeadersFetchDone( KJob *job )
{
    if ( job->error() ) {
        Warning() << job->errorString();
        emitResult();
        return;
    }
    emitResult();
}

QList< KMime::Message::Ptr > FetchMessagesJob::getMessages() const
{
    return mMessages.values();
}

KIMAP::MessageFlags FetchMessagesJob::getFlags(KMime::Message::Ptr msg) const
{
    return mFlags.value(mMessages.key(msg));
}

qint64 FetchMessagesJob::getImapUid(KMime::Message::Ptr msg) const
{
    return mUids.value(mMessages.key(msg));
}
