/*
 * Copyright (C) 2012  Christian Mollekopf <mollekopf@kolabsys.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "probekolabserverjob.h"
#include "findkolabfoldersjob.h"
#include "setupkolabfoldersjob.h"
#include <kimap/capabilitiesjob.h>
#include <kimap/namespacejob.h>
#include <errorhandler.h>

ProbeKolabServerJob::ProbeKolabServerJob(KIMAP::Session *session, QObject* parent)
:   KJob(parent),
    mSession(session)
{

}

ProbeKolabServerJob::~ProbeKolabServerJob()
{

}

void ProbeKolabServerJob::start()
{
    KIMAP::CapabilitiesJob *capabilities = new KIMAP::CapabilitiesJob(mSession);
    connect(capabilities, SIGNAL(result(KJob*)), this, SLOT(onCapabilitiesTestDone(KJob*)));
    capabilities->start();    
}

void ProbeKolabServerJob::createDefaultsIfMissing(const QStringList &folders)
{
    mFoldersToCreate = folders;
}

void ProbeKolabServerJob::onCapabilitiesTestDone(KJob* job)
{
    if ( job->error() ) {
        Warning() << job->errorString();
        setError(KJob::UserDefinedError);
        emitResult();
        return;
    }
    KIMAP::CapabilitiesJob *capabilitiesJob = qobject_cast<KIMAP::CapabilitiesJob*>( job );
    Q_ASSERT(capabilitiesJob);
    mCapabilities = capabilitiesJob->capabilities();
    if ( mCapabilities.contains( "NAMESPACE" ) ) {
        KIMAP::NamespaceJob *nsJob = new KIMAP::NamespaceJob( mSession );
        QObject::connect( nsJob, SIGNAL(result(KJob*)), SLOT(onNamespacesTestDone(KJob*)) );
        nsJob->start();
        return;
    } else {
        FindKolabFoldersJob *findJob = new FindKolabFoldersJob(mCapabilities, mPersonalNamespace, mExcludedNamespace, mSession, this);
        connect(findJob, SIGNAL(result(KJob*)), this, SLOT(findKolabFoldersDone(KJob*)));
        findJob->start();
    }
}


void ProbeKolabServerJob::onNamespacesTestDone( KJob *job )
{
    if ( job->error() ) {
        Warning() << job->errorString();
        setError(KJob::UserDefinedError);
        emitResult();
        return;
    }
    KIMAP::NamespaceJob *nsJob = qobject_cast<KIMAP::NamespaceJob*>( job );
    Q_ASSERT(nsJob);
    mPersonalNamespace = nsJob->personalNamespaces();
    mExcludedNamespace = nsJob->userNamespaces()+nsJob->sharedNamespaces();

    FindKolabFoldersJob *findJob = new FindKolabFoldersJob(mCapabilities, mPersonalNamespace, mExcludedNamespace, mSession, this);
    connect(findJob, SIGNAL(result(KJob*)), this, SLOT(findKolabFoldersDone(KJob*)));
    findJob->start();
}

void ProbeKolabServerJob::findKolabFoldersDone(KJob *job)
{
    if (job->error()) {
        Warning() << job->errorString();
        setError(KJob::UserDefinedError);
        emitResult();
        return;
    }
    FindKolabFoldersJob *findJob = static_cast<FindKolabFoldersJob*>(job);
//    qDebug() << "Found kolab folders: " << findJob->getKolabFolders();
    mKolabFolders = findJob->getKolabFolders();

    //TODO Setting up the missing folders shouldn't be part of the ProbeJob, and it's probably only used in one place in here => separate
    QString rootFolder;
    QStringList toCreate;
    foreach (const QString &folderType, mFoldersToCreate) {
        if (!mKolabFolders.contains(folderType)) {
            toCreate << folderType;
        }
    }
    if (toCreate.isEmpty()) {
        emitResult();
        return;
    }
    SetupKolabFoldersJob *setupJob = new SetupKolabFoldersJob(mCapabilities, rootFolder, mSession, this);
    setupJob->setKolabFolders(toCreate);
    connect(setupJob, SIGNAL(result(KJob*)), this, SLOT(onSetupDone(KJob*)));
    setupJob->start();
}

void ProbeKolabServerJob::onSetupDone(KJob *job)
{
    if (job->error()) {
        Warning() << job->errorString();
        setError(KJob::UserDefinedError);
    }
    SetupKolabFoldersJob *setupJob = static_cast<SetupKolabFoldersJob*>(job);
    const QMap<QString, QString> createdFolders = setupJob->createdFolders();
    foreach (const QString &folderType, createdFolders.keys()) {
        mKolabFolders.insert(folderType, createdFolders.value(folderType));
    }
    emitResult();
}


QMultiHash<QString, QString> ProbeKolabServerJob::kolabFolders() const
{
    return mKolabFolders;
}

QStringList ProbeKolabServerJob::capabilities() const
{
    return mCapabilities;
}

QList< KIMAP::MailBoxDescriptor > ProbeKolabServerJob::personalNamespace() const
{
    return mPersonalNamespace;
}

QList< KIMAP::MailBoxDescriptor > ProbeKolabServerJob::excludedNamespaces() const
{
    return mExcludedNamespace;
}
