/*
 * Copyright (C) 2012  Christian Mollekopf <mollekopf@kolabsys.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "kolabaccount.h"
#include "uiproxy.h"
#include <jobs/setupkolabfoldersjob.h>
#include "jobs/fetchmessagesjob.h"
#include <kimap/session.h>
#include <kimap/logoutjob.h>
#include <kimap/appendjob.h>
#include <kimap/createjob.h>
#include <kimap/deletejob.h>
#include <kimap/listjob.h>
#include <kimap/capabilitiesjob.h>
#include <kimap/namespacejob.h>
#include <kimap/storejob.h>
#include <kimap/selectjob.h>
#include <kimap/expungejob.h>
#include <kimap/setmetadatajob.h>
#include <kolab/errorhandler.h>
#include <kolab/kolabdefinitions.h>
#include <formathelpers.h>
#include <kolabobject.h>


KolabAccount::KolabAccount(QObject* parent)
:   QObject(parent),
    mSession(0),
    mEncryptionMode(KIMAP::LoginJob::TlsV1),
    mAuthenticationMode(KIMAP::LoginJob::Plain),
    mDryRun(false),
    mVersion(Kolab::KolabV3)
{

}

void KolabAccount::setVersion(Kolab::Version version)
{
    mVersion = version;
}

void KolabAccount::setHost(const QString& host, qint16 port)
{
    mHost = host;
    mPort = port;
}

void KolabAccount::setCredentials(const QString& username, const QString& pw, const QString &authorizationName)
{
    mUsername = username;
    mPw = pw;
    mAuthorizationName = authorizationName;
}

void KolabAccount::setDryRun(bool enable)
{
    mDryRun = enable;
}

bool KolabAccount::init()
{
    if (mSession) {
        return false;
    }
    mSession = new KIMAP::Session( mHost, mPort, this );
    mSession->setUiProxy( KIMAP::SessionUiProxy::Ptr(new UiProxy()) );
    KIMAP::LoginJob *loginJob = new KIMAP::LoginJob( mSession );
    Debug() << mUsername << mAuthorizationName << mPw;
    if (mAuthorizationName != mUsername) {
        loginJob->setAuthorizationName(mAuthorizationName);
    }
    loginJob->setUserName( mUsername );
    loginJob->setPassword( mPw );
    loginJob->setEncryptionMode( mEncryptionMode );
    loginJob->setAuthenticationMode( mAuthenticationMode );
    loginJob->exec();
    
    if ( loginJob->error() ) {
        Error() << "Failed to login: " << loginJob->errorString();
        mSession->close();
        return false;
    }
    Debug() << "authentication successful";
    
    KIMAP::CapabilitiesJob *capabilities = new KIMAP::CapabilitiesJob(mSession);
    capabilities->exec();
    mCapabilities = capabilities->capabilities();
    
    if ( mCapabilities.contains( "NAMESPACE" ) ) {
        KIMAP::NamespaceJob *nsJob = new KIMAP::NamespaceJob( mSession );
        nsJob->exec();
        mPersonalNamespaces = nsJob->personalNamespaces();
        mExcludedNamespaces = nsJob->userNamespaces();
        mExcludedNamespaces.append(nsJob->sharedNamespaces());
//         Debug() << "personal namespaces";
//         foreach (const KIMAP::MailBoxDescriptor &desc, nsJob->personalNamespaces()) {
//             Debug() << desc.name;
//         }
//         foreach (const KIMAP::MailBoxDescriptor &desc, nsJob->userNamespaces()) {
//             Debug() << desc.name;
//         }
//         foreach (const KIMAP::MailBoxDescriptor &desc, nsJob->sharedNamespaces()) {
//             Debug() << desc.name;
//         }
    }
    return true;
}

const char* FlagRecent2 = "\\Recent";



KMime::Message::Ptr KolabAccount::writeObject(const Object &obj)
{
    const QString productId("migration-utility");
    KMime::Message::Ptr msg;
    if (obj.object.canConvert<KMime::Message::Ptr>()) {
        msg = obj.object.value<KMime::Message::Ptr>();
    } else if (obj.object.canConvert<KCalCore::Incidence::Ptr>()) {
        msg = Kolab::KolabObjectWriter::writeIncidence(obj.object.value<KCalCore::Incidence::Ptr>(), mVersion, productId);
    } else if (obj.object.canConvert<KABC::Addressee>()) {
        msg = Kolab::KolabObjectWriter::writeContact(obj.object.value<KABC::Addressee>(), mVersion, productId);
    } else if (obj.object.canConvert<KABC::ContactGroup>()) {
        msg = Kolab::KolabObjectWriter::writeDistlist(obj.object.value<KABC::ContactGroup>(), mVersion, productId);
    } else if (obj.object.canConvert<Note>()) {
        msg = Kolab::KolabObjectWriter::writeNote(obj.object.value<Note>().msg, mVersion, productId);
    } else if (obj.object.canConvert<Dictionary>()) {
        const Dictionary dict = obj.object.value<Dictionary>();
        msg = Kolab::KolabObjectWriter::writeDictionary(dict.dict, dict.lang, mVersion, productId);
    } else {
        Error() << "Type not handled";
    }
    return msg;
}

void KolabAccount::appendObject(Object obj, const QString& folder)
{
    if (mDryRun) {
        Debug() << "append object in folder: " << folder;
        return;
    }
    Q_ASSERT(mFolders.contains(folder));

    KMime::Message::Ptr message = writeObject(obj);
    
    if (!message) {
        Error() << "got empty message";
        return;
    }
    KIMAP::AppendJob *job = new KIMAP::AppendJob( mSession );
    job->setMailBox( folder );
    job->setContent( message->encodedContent( true ) );
    //The Recent flag is a special case which is not allowed in the append command
    obj.flags.removeAll(QByteArray(FlagRecent2));
    job->setFlags( obj.flags );
    job->exec();
    Debug() << "appended object in folder: " << folder;
    if (job->error()) {
        Error() << job->errorString();
    }
}

void KolabAccount::logout()
{
    Q_ASSERT(mSession);
    KIMAP::LogoutJob *logoutJob = new KIMAP::LogoutJob(mSession);
    logoutJob->exec();
    mSession->close();
//     mSession->deleteLater();
    mSession = 0;
    Debug() << "logout done";
}

const char* FlagDeleted2 = "\\Deleted";

void KolabAccount::cleanAccount()
{
    Q_ASSERT(mSession);
    KIMAP::ListJob *listJob = new KIMAP::ListJob(mSession);
    listJob->setOption(KIMAP::ListJob::IncludeUnsubscribed);
    listJob->setQueriedNamespaces(mPersonalNamespaces);
    QObject::connect( listJob, SIGNAL(mailBoxesReceived(QList<KIMAP::MailBoxDescriptor>,QList<QList<QByteArray> >)),
                      this, SLOT(mailBoxesReceived(QList<KIMAP::MailBoxDescriptor>,QList<QList<QByteArray> >)));
    listJob->exec();
    if (listJob->error()) {
        Error() << listJob->errorString();
    }
    foreach (const KIMAP::MailBoxDescriptor &desc, mMailboxes) {
        bool skip = false;
        foreach(const KIMAP::MailBoxDescriptor &excluded, mExcludedNamespaces) {
            if (desc.name.contains(excluded.name)) {
                skip = true;
                break;
            }
        }
        if (skip) {
            continue;
        }

        KIMAP::SelectJob *selectJob = new KIMAP::SelectJob(mSession);
        selectJob->setMailBox(desc.name);
        selectJob->exec();
        if (selectJob->error()) {
            Error() << selectJob->errorString();
        }
        const int messageCount = selectJob->messageCount();
        if (mDryRun) {
            Debug() << "removing folder: " << desc.name << messageCount;
            continue;
        }

        if (messageCount > 0) {
            KIMAP::StoreJob *storeJob = new KIMAP::StoreJob(mSession);
            storeJob->setMode(KIMAP::StoreJob::AppendFlags);
            storeJob->setFlags(KIMAP::MessageFlags() << FlagDeleted2);
            storeJob->setSequenceSet( KIMAP::ImapSet( 1, messageCount ) );
            storeJob->exec();
            if (storeJob->error()) {
                Error() << storeJob->errorString();
            }

            KIMAP::ExpungeJob *expungeJob = new KIMAP::ExpungeJob(mSession);
            expungeJob->exec();
            if (expungeJob->error()) {
                Error() << expungeJob->errorString();
            }
        }
        if (desc.name.contains(QLatin1String("inbox"), Qt::CaseInsensitive)) {
            continue;
        }
        
        KIMAP::DeleteJob *deleteJob = new KIMAP::DeleteJob(mSession);
        Debug() << "deleting: " << desc.name;
        deleteJob->setMailBox(desc.name);
        deleteJob->exec();
        if (deleteJob->error()) {
            Error() << deleteJob->errorString();
        }
    }
    mFolders.clear();
    mFolders.append(QLatin1String("inbox"));

    //logging out and logging in again seems to help with the "NO Mailbox is locked" problem when creating folders afterwards.
    logout();
    init();
}

void KolabAccount::mailBoxesReceived(const QList<KIMAP::MailBoxDescriptor> &descriptors, const QList< QList< QByteArray > > &flags)
{
    mMailboxes.append(descriptors);
}

//TODO move to formathelpers.h or kolabdefinitions.h
#define KOLAB_FOLDER_TYPE_DRAFTS_SUFFIX ".drafts"
#define KOLAB_FOLDER_TYPE_SENT_SUFFIX ".sentitems"
#define KOLAB_FOLDER_TYPE_OUTBOX_SUFFIX ".outbox"
#define KOLAB_FOLDER_TYPE_TRASH_SUFFIX ".wastebasket"
#define KOLAB_FOLDER_TYPE_JUNK_SUFFIX ".junkemail"


/*
 * This can fail with "NO Mailbox is locked" if the we deleted a selected folder before:
 * http://lists.andrew.cmu.edu/pipermail/info-cyrus/2011-June/035049.html
 * 
 */
void KolabAccount::createFolder(const QString &name, const QByteArray &annotation)
{
    if (mFolders.contains(name, Qt::CaseInsensitive)) {
        Warning() << "folder is already existing: " << name;
        return;
    }
    if (mDryRun) {
        Debug() << "creating folder: " << name << annotation;
        return;
    }

    KIMAP::CreateJob *createJob = new KIMAP::CreateJob(mSession);
    createJob->setMailBox(name);
    createJob->exec();
    Debug() << "created folder " << name;
    if (createJob->error()) {
        Error() << createJob->errorString();
        return;
    }
    
    KIMAP::SetMetaDataJob *setMetadataJob = new KIMAP::SetMetaDataJob(mSession);
    setMetadataJob->setMailBox(createJob->mailBox());
    if ( mCapabilities.contains( "METADATA" ) ) {
        setMetadataJob->setServerCapability( KIMAP::MetaDataJobBase::Metadata );
        setMetadataJob->addMetaData( "/shared" KOLAB_FOLDER_TYPE_ANNOTATION, annotation );
    } else {
        setMetadataJob->setServerCapability( KIMAP::MetaDataJobBase::Annotatemore );
        setMetadataJob->setEntry( KOLAB_FOLDER_TYPE_ANNOTATION );
        setMetadataJob->addMetaData( "value.shared", annotation );
    }
    setMetadataJob->exec();
    if (setMetadataJob->error()) {
        Error() << setMetadataJob->errorString();
    }

    mFolders.append(name);
}

void KolabAccount::createFolder(const QString& name, Kolab::FolderType folderType)
{
    createFolder(name, QString::fromStdString(Kolab::folderAnnotation(folderType)).toLatin1());
}

void KolabAccount::setupFolders()
{
    if (mDryRun) {
        Debug() << "setup default folders";
        return;
    }
    SetupKolabFoldersJob *setupJob = new SetupKolabFoldersJob(mCapabilities, QString(), mSession, this);
    setupJob->setKolabFolders(QStringList() << KOLAB_FOLDER_TYPE_CONTACT << KOLAB_FOLDER_TYPE_EVENT << KOLAB_FOLDER_TYPE_TASK << KOLAB_FOLDER_TYPE_JOURNAL << KOLAB_FOLDER_TYPE_NOTE);
    setupJob->exec();
    if (setupJob->error()) {
        Error() << setupJob->errorString();
    }
    foreach (const QString &folder, setupJob->createdFolders().values()) {
        mFolders.append(folder);
    }
    KIMAP::SelectJob *selectJob = new KIMAP::SelectJob(mSession);
    selectJob->setMailBox(QString());
    //TODO internationalize folder names
    //TODO make part of setupkolabfoldersjob after adding types to folderTypeFromString in formathelpers.cpp
    //TODO the suffixes should go in the private namespace
    createFolder("Configuration", KOLAB_FOLDER_TYPE_CONFIGURATION KOLAB_FOLDER_TYPE_DEFAULT_SUFFIX);
    createFolder("Drafts", KOLAB_FOLDER_TYPE_MAIL KOLAB_FOLDER_TYPE_DRAFTS_SUFFIX);
    createFolder("Sent", KOLAB_FOLDER_TYPE_MAIL KOLAB_FOLDER_TYPE_SENT_SUFFIX);
    createFolder("Trash", KOLAB_FOLDER_TYPE_MAIL KOLAB_FOLDER_TYPE_TRASH_SUFFIX);
}

QList<Object> KolabAccount::getObjects(const QString& folder)
{
//     Debug() << folder;
    FetchMessagesJob *fetchJob = new FetchMessagesJob(folder, mSession, this);
    fetchJob->exec();
    Debug() << fetchJob->getMessages().size();
    QList<Object> messages;
    foreach (const KMime::Message::Ptr &msg, fetchJob->getMessages()) {
        Object obj;
        obj.object = QVariant::fromValue(msg);
        obj.flags = fetchJob->getFlags(msg);
        messages.append(obj);
    }
    return messages;
}

QStringList KolabAccount::lookupFolderList()
{
//     Debug() << "lookupFolderList" << mHost << mPort << mUsername << mPw;
    init();
    mMailboxes.clear();
    
    KIMAP::ListJob *listJob = new KIMAP::ListJob(mSession);
    listJob->setOption(KIMAP::ListJob::IncludeUnsubscribed);
    listJob->setQueriedNamespaces(mPersonalNamespaces);
    QObject::connect( listJob, SIGNAL(mailBoxesReceived(QList<KIMAP::MailBoxDescriptor>,QList<QList<QByteArray> >)),
                      this, SLOT(mailBoxesReceived(QList<KIMAP::MailBoxDescriptor>,QList<QList<QByteArray> >)));
    listJob->exec();
    Debug() << "found " << mMailboxes.size();
    
    QStringList mailboxes;
    foreach (const KIMAP::MailBoxDescriptor &descriptor, mMailboxes) {
        mailboxes.append(descriptor.name);
    }
    return mailboxes;
}

