/*
 * Copyright (C) 2012  Christian Mollekopf <mollekopf@kolabsys.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <QtCore/qcoreapplication.h>
#include <QtCore/QStringList>
#include <QtCore/qfile.h>
#include <kcmdlineargs.h>
#include <kolab/errorhandler.h>

#include "coordinationjob.h"
#include "sourceserver.h"
#include "kolabserver.h"

KIMAP::LoginJob::EncryptionMode getEncryptionMode(const QString &encrypt)
{
    KIMAP::LoginJob::EncryptionMode encryptionMode = KIMAP::LoginJob::TlsV1;
    if (encrypt == "NONE") {
        encryptionMode = KIMAP::LoginJob::Unencrypted;
    } else if (encrypt == "SSL") {
        encryptionMode = KIMAP::LoginJob::AnySslVersion;
    } else if (encrypt == "TLS") {
        encryptionMode = KIMAP::LoginJob::TlsV1;
    } else {
        Error() << "unknown encryption mode";
    }
    return encryptionMode;
}

KIMAP::LoginJob::AuthenticationMode getAuthenticationMode(const QString &auth)
{
    KIMAP::LoginJob::AuthenticationMode authenticationMode = KIMAP::LoginJob::Plain;
    if (auth == "PLAIN") {
        authenticationMode = KIMAP::LoginJob::Plain;
    } else if (auth == "LOGIN") {
        authenticationMode = KIMAP::LoginJob::Login;
    } else if (auth == "CRAMMD5") {
        authenticationMode = KIMAP::LoginJob::CramMD5;
    } else if (auth == "DIGESTMD5") {
        authenticationMode = KIMAP::LoginJob::DigestMD5;
    } else if (auth == "NTLM") {
        authenticationMode = KIMAP::LoginJob::NTLM;
    } else if (auth == "GSSAPI") {
        authenticationMode = KIMAP::LoginJob::GSSAPI;
    } else if (auth == "ANONYMOUS") {
        authenticationMode = KIMAP::LoginJob::Anonymous;
    } else if (auth == "CLEARTEXT") {
        authenticationMode = KIMAP::LoginJob::ClearText;
    } else {
        Error() << "unknown authentication mode";
    }
    return authenticationMode;
}

/**
 * kolab-migrate --from caldav --to kolab (implied?) --caldav-host caldav.example.org --caldav-authn <admin> --caldav-authz <user> --caldav-pass <pass> --<folder-selection> --<folder-mapping> --kolab-host kolab.example.org --kolab-authn <admin> --kolab-authz <user> --kolab-pass <pass>

* --regextrans2 '/Public Folders/shared/' Regex transformation of folder paths
* Incremental conversion (folder by folder or even by batch of messages)
* Save state persistently and resume from there after interruption (Wipe before starting and start all over as first step)
 */

int main(int argc, char *argv[])
{
    KCmdLineArgs::init(argc, argv, "migrationutility", "migrationutility",ki18n("migrationutility"),  "0.1");
    
    KCmdLineOptions options;
    options.add("dry", ki18n("Dry run, doesn't change anything on the server, but only prints what would be done (Use with care until this is well tested)."));
    options.add("from <type>", ki18n("Source host type (kolab2, kolab3, exchangeimap)"));
    
    options.add("from-kolab2-host <host>", ki18n("Source Kolab2 host"));
    options.add("from-kolab2-port <port>", ki18n("Port"), "143");
    options.add("from-kolab2-user <loginname>", ki18n("Username for single-user migration"));
    options.add("from-kolab2-proxyauth <loginname>", ki18n("Username to be used for authentication together with password (optional, works with PLAIN/SASL authentication)"));
    options.add("from-kolab2-password <password>", ki18n("Password"));
    options.add("from-kolab2-encrypt <mode>", ki18n("Encryption mode to be used (NONE, TLS, SSL)"), "TLS");
    options.add("from-kolab2-auth <mode>", ki18n("Authentication mode to be used (PLAIN, LOGIN, CRAMMD5, DIGESTMD5, NTLM, GSSAPI, ANONYMOUS, CLEARTEXT)"), "PLAIN");
    
    options.add("from-kolab3-host <host>", ki18n("Source Kolab3 host"));
    options.add("from-kolab3-port <port>", ki18n("Port"), "143");
    options.add("from-kolab3-user <loginname>", ki18n("Username for single-user migration"));
    options.add("from-kolab3-proxyauth <loginname>", ki18n("Username to be used for authentication together with password (optional, works with PLAIN/SASL authentication)"));
    options.add("from-kolab3-password <password>", ki18n("Password"));
    options.add("from-kolab3-encrypt <mode>", ki18n("Encryption mode to be used (NONE, TLS, SSL)"), "TLS");
    options.add("from-kolab3-auth <mode>", ki18n("Authentication mode to be used (PLAIN, LOGIN, CRAMMD5, DIGESTMD5, NTLM, GSSAPI, ANONYMOUS, CLEARTEXT)"), "PLAIN");
    
    options.add("from-exchangeimap-host <host>", ki18n("Source Exchange IMAP host"));
    options.add("from-exchangeimap-port <port>", ki18n("Port"), "143");
    options.add("from-exchangeimap-user <loginname>", ki18n("Username for single-user migration"));
    options.add("from-exchangeimap-proxyauth <loginname>", ki18n("Username to be used for authentication together with password (optional, works with PLAIN/SASL authentication)"));
    options.add("from-exchangeimap-password <password>", ki18n("Password"));
    options.add("from-exchangeimap-encrypt <mode>", ki18n("Encryption mode to be used (NONE, TLS, SSL)"), "TLS");
    options.add("from-exchangeimap-auth <mode>", ki18n("Authentication mode to be used (PLAIN, LOGIN, CRAMMD5, DIGESTMD5, NTLM, GSSAPI, ANONYMOUS, CLEARTEXT)"), "PLAIN");
    
    options.add("to <type>", ki18n("Target host type (kolab2, kolab3)"));
    
    options.add("to-kolab2-host <host>", ki18n("Target Kolab2 host"));
    options.add("to-kolab2-port <port>", ki18n("Port"), "143");
    options.add("to-kolab2-user <loginname>", ki18n("Username for single-user migration"));
    options.add("to-kolab2-proxyauth <loginname>", ki18n("Username to be used for authentication together with password (optional, works with PLAIN/SASL authentication)"));
    options.add("to-kolab2-password <password>", ki18n("Password"));

    options.add("to-kolab3-host <host>", ki18n("Target Kolab3 host"));
    options.add("to-kolab3-port <port>", ki18n("Port"), "143");
    options.add("to-kolab3-user <loginname>", ki18n("Username for single-user migration"));
    options.add("to-kolab3-proxyauth <loginname>", ki18n("Username to be used for authentication together with password (optional, works with PLAIN/SASL authentication)"));
    options.add("to-kolab3-password <password>", ki18n("Password"));
    
    KCmdLineArgs::addCmdLineOptions( options );
    QCoreApplication app(argc, argv);
    
    KCmdLineArgs *args = KCmdLineArgs::parsedArgs();

    SourceServer *sourceServer = 0;
    if (!args->getOption("from").compare("kolab2", Qt::CaseInsensitive)) {
        Debug() << "From Kolab2 Server";
        KolabSourceServer *kolabSourceServer = new KolabSourceServer(&app);
        kolabSourceServer->setHost(args->getOption("from-kolab2-host"), args->getOption("from-kolab2-port").toInt());
        kolabSourceServer->setAdminCredentials(args->getOption("from-kolab2-proxyauth"), args->getOption("from-kolab2-password"));
        kolabSourceServer->setEncryption(getEncryptionMode(args->getOption("from-kolab2-encrypt")));
        kolabSourceServer->setAuthentication(getAuthenticationMode(args->getOption("from-kolab2-auth")));
        if (args->isSet("from-kolab2-user")) {
            kolabSourceServer->setSingleUser(args->getOption("from-kolab2-user"));
        }
        sourceServer = kolabSourceServer;
    } else if (!args->getOption("from").compare("kolab3", Qt::CaseInsensitive)) {
        Debug() << "From Kolab3 Server";
        KolabSourceServer *kolabSourceServer = new KolabSourceServer(&app);
        kolabSourceServer->setHost(args->getOption("from-kolab3-host"), args->getOption("from-kolab3-port").toInt());
        kolabSourceServer->setAdminCredentials(args->getOption("from-kolab3-proxyauth"), args->getOption("from-kolab3-password"));
        kolabSourceServer->setEncryption(getEncryptionMode(args->getOption("from-kolab3-encrypt")));
        kolabSourceServer->setAuthentication(getAuthenticationMode(args->getOption("from-kolab3-auth")));
        if (args->isSet("from-kolab3-user")) {
            kolabSourceServer->setSingleUser(args->getOption("from-kolab3-user"));
        }
        sourceServer = kolabSourceServer;
    } else if (!args->getOption("from").compare("exchangeimap", Qt::CaseInsensitive)) {
        Debug() << "From Exchange IMAP Server";
        ExchangeIMAPSourceServer *source = new ExchangeIMAPSourceServer(&app);
        source->setHost(args->getOption("from-exchangeimap-host"), args->getOption("from-exchangeimap-port").toInt());
        source->setAdminCredentials(args->getOption("from-exchangeimap-proxyauth"), args->getOption("from-exchangeimap-password"));
        source->setEncryption(getEncryptionMode(args->getOption("from-exchangeimap-encrypt")));
        source->setAuthentication(getAuthenticationMode(args->getOption("from-exchangeimap-auth")));
        if (args->isSet("from-exchangeimap-user")) {
            source->setSingleUser(args->getOption("from-exchangeimap-user"));
        }
        sourceServer = source;
    }

    KolabServer *kolabServer = 0;
    if (!args->getOption("to").compare("kolab2", Qt::CaseInsensitive)) {
        Debug() << "To Kolab2 Server";
        kolabServer = new KolabServer(&app);
        kolabServer->setHost(args->getOption("to-kolab2-host"), args->getOption("to-kolab2-port").toInt());
        kolabServer->setAdminCredentials(args->getOption("to-kolab2-proxyauth"), args->getOption("to-kolab2-password"));
        kolabServer->setVersion(Kolab::KolabV2);
    } else if (!args->getOption("to").compare("kolab3", Qt::CaseInsensitive)) {
        Debug() << "To Kolab3 Server";
        kolabServer = new KolabServer(&app);
        kolabServer->setHost(args->getOption("to-kolab3-host"), args->getOption("to-kolab3-port").toInt());
        kolabServer->setAdminCredentials(args->getOption("to-kolab3-proxyauth"), args->getOption("to-kolab3-password"));
        kolabServer->setVersion(Kolab::KolabV3);
    }

    if (!sourceServer || !kolabServer) {
        Error() << "Target or Source-Server not specified";
        return -1;
    }
    if (args->isSet("dry")) {
        kolabServer->setDryRun(true);
    }
    CoordinationJob *job = new CoordinationJob(sourceServer, kolabServer, &app);
    QObject::connect(job, SIGNAL(done()), &app, SLOT(quit()));
    job->start();
    //TODO reenable once we have async jobs which don't execute everything using exec
    //app.exec();
    
    if (Kolab::ErrorHandler::instance().error() >= Kolab::ErrorHandler::Error) {
        Error() << Kolab::ErrorHandler::instance().errorMessage();
        return -1;
    }
    return 0;
}
