/*
 * Copyright (C) 2012  Christian Mollekopf <mollekopf@kolabsys.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "sourceserver.h"
#include "uiproxy.h"
#include <jobs/getuserlistjob.h>
#include <kimap/listjob.h>
#include <kimap/logoutjob.h>
#include <kolab/commonconversion.h>
#include <kolab/errorhandler.h>


SourceServer::SourceServer(QObject* parent): QObject(parent)
{

}

SourceAccount* SourceServer::getSourceAccount(const QString& user)
{
    return 0;
}


QStringList SourceServer::getUserList()
{
    if (!mExplicitUsers.isEmpty()) {
        return mExplicitUsers;
    }
    return retrieveUserList();
}

void SourceServer::setSingleUser(const QString &user)
{
    mExplicitUsers.append(user);
}

QStringList SourceServer::retrieveUserList()
{
    return QStringList();
}




TestServer::TestServer(QObject* parent)
:   SourceServer(parent)
{
    mUsers << QLatin1String("user1");
    mUsers << QLatin1String("user2");
    mUsers << QLatin1String("user3");
}

QStringList TestServer::retrieveUserList()
{
    return mUsers;
}

SourceAccount* TestServer::getSourceAccount(const QString& user)
{
    Q_ASSERT(mUsers.removeAll(user) == 1);
    return new TestAccount(this);
}


//---------------------------------------------------------------------------

KolabSourceServer::KolabSourceServer(QObject* parent)
:   SourceServer(parent),
    mSession(0),
    mEncryptionMode(KIMAP::LoginJob::TlsV1),
    mAuthenticationMode(KIMAP::LoginJob::Plain)
{

}

void KolabSourceServer::setHost(const QString& host, qint16 port)
{
    mHost = host;
    mPort = port;
}

void KolabSourceServer::setAdminCredentials(const QString& username, const QString& pw)
{
    mUsername = username;
    mPw = pw;
}

void KolabSourceServer::setEncryption(KIMAP::LoginJob::EncryptionMode enc)
{
    mEncryptionMode = enc;
}

void KolabSourceServer::setAuthentication(KIMAP::LoginJob::AuthenticationMode auth)
{
    mAuthenticationMode = auth;
}

SourceAccount* KolabSourceServer::getSourceAccount(const QString& user)
{
    KIMAP::Session *session = new KIMAP::Session( mHost, mPort, this );
    session->setUiProxy( KIMAP::SessionUiProxy::Ptr(new UiProxy()) );
    KIMAP::LoginJob *loginJob = new KIMAP::LoginJob( session );
    if (user != mUsername) {
        loginJob->setAuthorizationName(mUsername);
    }
    loginJob->setUserName( user );
    loginJob->setPassword( mPw );
    loginJob->setEncryptionMode( mEncryptionMode );
    loginJob->setAuthenticationMode( mAuthenticationMode );
    loginJob->exec();

    if ( loginJob->error() ) {
        Error() << "Failed to login: " << loginJob->errorString();
        return 0;
    } else {
        Debug() << "authentication successful";
    }
    
    return new KolabSourceAccount(session, this);
}

QStringList KolabSourceServer::retrieveUserList()
{
    mSession = new KIMAP::Session( mHost, mPort, this );
    mSession->setUiProxy( KIMAP::SessionUiProxy::Ptr(new UiProxy()) );
    
    KIMAP::LoginJob *loginJob = new KIMAP::LoginJob( mSession );
    loginJob->setUserName( mUsername );
    loginJob->setPassword( mPw );
    loginJob->setEncryptionMode( mEncryptionMode );
    loginJob->setAuthenticationMode( mAuthenticationMode );
    loginJob->exec();
    
    if ( loginJob->error() ) {
        Warning() << "Failed to login: " << loginJob->errorString();
    } else {
        Debug() << "authentication successful";
    }
    
    GetUserListJob *userListJob = new GetUserListJob(mSession, this);
    userListJob->exec();
    const QStringList userList = userListJob->getUserList();
    logout();

    return userList;
}


void KolabSourceServer::logout()
{
    Q_ASSERT(mSession);
    KIMAP::LogoutJob *logoutJob = new KIMAP::LogoutJob(mSession);
    logoutJob->exec();
    mSession->close();
    mSession->deleteLater();
    mSession = 0;
    Debug() << "logout done";
}

//---------------------------------------------------------------------

ExchangeIMAPSourceServer::ExchangeIMAPSourceServer(QObject* parent)
:   KolabSourceServer(parent)
{

}

SourceAccount* ExchangeIMAPSourceServer::getSourceAccount(const QString& user)
{
    KIMAP::Session *session = new KIMAP::Session( mHost, mPort, this );
    session->setUiProxy( KIMAP::SessionUiProxy::Ptr(new UiProxy()) );
    KIMAP::LoginJob *loginJob = new KIMAP::LoginJob( session );
    if (user != mUsername) {
        loginJob->setAuthorizationName(mUsername);
    }
    loginJob->setUserName( user );
    loginJob->setPassword( mPw );
    loginJob->setEncryptionMode( mEncryptionMode );
    loginJob->setAuthenticationMode( mAuthenticationMode );
    loginJob->exec();

    if ( loginJob->error() ) {
        Error() << "Failed to login: " << loginJob->errorString();
        return 0;
    } else {
        Debug() << "authentication successful";
    }
    
    return new ExchangeIMAPSourceAccount(session, this);
}